from decimal import Decimal

from dateutil.relativedelta import relativedelta
from django import forms
from django.test import TestCase
from edc_utils import get_utcnow

from edc_vitals.form_validators import WeightHeightBmiFormValidatorMixin

from ..models import WeightHeightBmi


class TestWeightHeightBmi(TestCase):
    def test_allows_none(self):
        obj = WeightHeightBmi()
        obj.save()

        obj = WeightHeightBmi(weight=65.0)
        obj.save()

        obj = WeightHeightBmi(height=65.0)
        obj.save()

    def test_calculates_bmi(self):
        obj = WeightHeightBmi(
            weight=Decimal("65.0"),
            height=Decimal("180.0"),
            dob=get_utcnow() - relativedelta(years=25),
        )
        obj.save()
        self.assertEqual(round(obj.calculated_bmi_value, 4), 20.0617)

    def test_form_validator(self):
        form_validator = WeightHeightBmiFormValidatorMixin()
        form_validator.validate_weight_height_with_bmi()
        cleaned_data = dict(weight=65, height=180)
        form_validator.validate_weight_height_with_bmi(
            weight_kg=cleaned_data.get("weight"),
            height_cm=cleaned_data.get("height"),
        )

        cleaned_data = dict(weight=65, height=18)
        self.assertRaises(
            forms.ValidationError,
            form_validator.validate_weight_height_with_bmi,
            weight_kg=cleaned_data.get("weight"),
            height_cm=cleaned_data.get("height"),
        )

        cleaned_data = dict(weight=65, height=180)
        self.assertRaises(
            forms.ValidationError,
            form_validator.validate_weight_height_with_bmi,
            weight_kg=cleaned_data.get("weight"),
            height_cm=cleaned_data.get("height"),
            dob=get_utcnow() - relativedelta(years=10),
        )

        cleaned_data = dict(weight=65, height=180)
        try:
            form_validator.validate_weight_height_with_bmi(
                weight_kg=cleaned_data.get("weight"),
                height_cm=cleaned_data.get("height"),
                dob=get_utcnow() - relativedelta(years=18),
            )
        except forms.ValidationError as e:
            self.fail(f"forms.ValidationError unexpectedly raised. Got {e}")



from django.db import models
from edc_utils import get_utcnow

from edc_vitals.model_mixins import (
    BloodPressureModelMixin,
    SimpleBloodPressureModelMixin,
    WeightHeightBmiModelMixin,
)


class BloodPressure(BloodPressureModelMixin, models.Model):
    pass


class SimpleBloodPressure(SimpleBloodPressureModelMixin, models.Model):
    pass


class WeightHeightBmi(WeightHeightBmiModelMixin, models.Model):

    report_datetime = models.DateTimeField(default=get_utcnow)

    dob = models.DateField(null=True)



from django import forms
from edc_reportable import CalculatorError, calculate_bmi


class WeightHeightBmiFormValidatorMixin:
    @staticmethod
    def validate_weight_height_with_bmi(weight_kg=None, height_cm=None, **kwargs):
        try:
            bmi = calculate_bmi(weight_kg=weight_kg, height_cm=height_cm, **kwargs)
        except CalculatorError as e:
            raise forms.ValidationError(f"Please check weight and height. {e}")
        return bmi



from django import forms
from django.test import TestCase
from edc_constants.constants import YES

from edc_vitals.form_validators import BloodPressureFormValidatorMixin
from edc_vitals.utils import has_severe_htn

from ..models import BloodPressure, SimpleBloodPressure


class TestBloodPressure(TestCase):
    def test_allows_none(self):
        obj = BloodPressure()
        obj.save()

    def test_calculates_avg_only_if_fields_provided(self):
        obj = BloodPressure(sys_blood_pressure_one=100)
        obj.save()
        obj.refresh_from_db()
        self.assertIsNone(obj.sys_blood_pressure_avg, obj.dia_blood_pressure_avg)

        obj = BloodPressure(sys_blood_pressure_one=100, dia_blood_pressure_one=100)
        obj.save()
        obj.refresh_from_db()
        self.assertIsNone(obj.sys_blood_pressure_avg, obj.dia_blood_pressure_avg)

        obj = BloodPressure(
            sys_blood_pressure_one=100,
            dia_blood_pressure_one=100,
            sys_blood_pressure_two=100,
            dia_blood_pressure_two=100,
        )
        obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.sys_blood_pressure_avg, 100)
        self.assertEqual(obj.dia_blood_pressure_avg, 100)

    def test_calculates_avg(self):
        obj = BloodPressure(
            sys_blood_pressure_one=110,
            dia_blood_pressure_one=80,
            sys_blood_pressure_two=128,
            dia_blood_pressure_two=91,
        )
        obj.save()
        obj.refresh_from_db()
        self.assertEqual(obj.sys_blood_pressure_avg, 119)
        self.assertEqual(obj.dia_blood_pressure_avg, 85)

    def test_simple_allows_none(self):
        obj = SimpleBloodPressure()
        obj.save()

    def test_simple_ok(self):
        obj = SimpleBloodPressure(
            sys_blood_pressure=110,
            dia_blood_pressure=80,
        )
        obj.save()

    def test_has_severe_htn(self):
        self.assertIsNone(has_severe_htn())
        self.assertIsNone(has_severe_htn(sys=120))
        self.assertFalse(has_severe_htn(sys=120, dia=80))
        self.assertFalse(has_severe_htn(sys=179, dia=109))
        self.assertTrue(has_severe_htn(sys=180, dia=109))
        self.assertTrue(has_severe_htn(sys=179, dia=110))
        self.assertTrue(has_severe_htn(sys=180, dia=110))

    def test_bp_suggest_severe_htn_form_validator(self):
        cleaned_data = dict(
            sys_blood_pressure_one=None,
            sys_blood_pressure_two=None,
            dia_blood_pressure_one=None,
            dia_blood_pressure_two=None,
        )
        form_validator = BloodPressureFormValidatorMixin()
        self.assertIsNone(
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn(**cleaned_data)
        )
        cleaned_data.update(sys_blood_pressure_one=180, dia_blood_pressure_one=None)
        self.assertIsNone(
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn(**cleaned_data)
        )
        cleaned_data.update(sys_blood_pressure_one=180, dia_blood_pressure_one=120)
        self.assertIsNone(
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn(**cleaned_data)
        )
        cleaned_data.update(
            sys_blood_pressure_one=180,
            dia_blood_pressure_one=120,
            sys_blood_pressure_two=191,
            dia_blood_pressure_two=125,
        )
        self.assertRaises(
            forms.ValidationError,
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn,
            **cleaned_data,
        )

        cleaned_data.update(severe_htn=YES)
        try:
            form_validator.raise_on_avg_blood_pressure_suggests_severe_htn(**cleaned_data)
        except forms.ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models


class WeightField(models.DecimalField):

    description = "Weight in Kg"

    def __init__(self, *args, **kwargs):
        if not kwargs.get("verbose_name"):
            kwargs["verbose_name"] = "Weight:"
        if not kwargs.get("validators"):
            kwargs["validators"] = [MinValueValidator(15), MaxValueValidator(200)]
        if not kwargs.get("help_text"):
            kwargs["help_text"] = "in kg"
        kwargs["max_digits"] = 8
        kwargs["decimal_places"] = 2
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["verbose_name"]
        del kwargs["max_digits"]
        del kwargs["decimal_places"]
        del kwargs["validators"]
        del kwargs["help_text"]
        return name, path, args, kwargs



from django.db import models
from edc_constants.choices import YES_NO

from ..models import DiastolicPressureField, SystolicPressureField
from ..utils import calculate_avg_bp


class SimpleBloodPressureModelMixin(models.Model):

    sys_blood_pressure = SystolicPressureField(
        null=True,
        blank=True,
    )

    dia_blood_pressure = DiastolicPressureField(
        null=True,
        blank=True,
    )

    class Meta:
        abstract = True


class BloodPressureModelMixin(models.Model):
    sys_blood_pressure_one = SystolicPressureField(
        verbose_name="Blood pressure: systolic (first reading)",
        null=True,
        blank=True,
    )

    dia_blood_pressure_one = DiastolicPressureField(
        verbose_name="Blood pressure: diastolic (first reading)",
        null=True,
        blank=True,
    )

    sys_blood_pressure_two = SystolicPressureField(
        verbose_name="Blood pressure: systolic (second reading)",
        null=True,
        blank=True,
    )

    dia_blood_pressure_two = DiastolicPressureField(
        verbose_name="Blood pressure: diastolic (second reading)",
        null=True,
        blank=True,
    )

    sys_blood_pressure_avg = models.IntegerField(
        verbose_name="Blood pressure: systolic (average)",
        null=True,
        blank=True,
    )

    dia_blood_pressure_avg = models.IntegerField(
        verbose_name="Blood pressure: diastolic (average)",
        null=True,
        blank=True,
    )

    # TODO: is this being validated

    severe_htn = models.CharField(
        verbose_name="Does the patient have severe hypertension?",
        max_length=15,
        choices=YES_NO,
        help_text="Based on the above readings. Severe HTN is any BP reading > 180/110mmHg",
        null=True,
        blank=True,
    )

    def save(self, *args, **kwargs):
        (
            self.sys_blood_pressure_avg,
            self.dia_blood_pressure_avg,
        ) = calculate_avg_bp(**self.__dict__)
        super().save(*args, **kwargs)

    class Meta:
        abstract = True



from django import forms
from edc_constants.constants import YES

from ..utils import calculate_avg_bp, has_severe_htn


class BloodPressureFormValidatorMixin:
    """Coupled with BloodPressureModelMixin"""

    @staticmethod
    def raise_on_avg_blood_pressure_suggests_severe_htn(
        use_avg=None, severe_htn_field_name=None, errmsg=None, **kwargs
    ):
        """Raise if BP is >= 180/110, See settings"""
        severe_htn_field_name = severe_htn_field_name or "severe_htn"
        severe_htn_reponse = kwargs.get(severe_htn_field_name)
        errmsg = {
            severe_htn_field_name: (errmsg or "Invalid. Patient has severe hypertension")
        }
        use_avg = True if use_avg is None else False
        avg_sys, avg_dia = calculate_avg_bp(use_av=use_avg, **kwargs)
        if has_severe_htn(sys=avg_sys, dia=avg_dia):
            if severe_htn_reponse != YES:
                raise forms.ValidationError(errmsg)



from django.conf import settings


def calculate_avg_bp(
    sys_blood_pressure_one=None,
    sys_blood_pressure_two=None,
    dia_blood_pressure_one=None,
    dia_blood_pressure_two=None,
    **kwargs
):
    avg_sys = None
    avg_dia = None
    if sys_blood_pressure_one and sys_blood_pressure_two:
        avg_sys = (sys_blood_pressure_one + sys_blood_pressure_two) / 2
    if dia_blood_pressure_one and dia_blood_pressure_two:
        avg_dia = (dia_blood_pressure_one + dia_blood_pressure_two) / 2
    return (avg_sys, avg_dia) if avg_sys and avg_dia else (None, None)


def has_severe_htn(sys=None, dia=None):
    if sys is not None and dia is not None:
        return sys >= get_sys_upper() or dia >= get_dia_upper()
    return None


def get_sys_upper():
    return getattr(settings, "EDC_VITALS_SYS_UPPER", 180)


def get_dia_upper():
    return getattr(settings, "EDC_VITALS_DIA_UPPER", 110)
